"""Stored command arguments for plugins."""
import shlex
from typing import Optional

from meltano.core.behavior.canonical import Canonical
from meltano.core.error import Error
from meltano.core.utils import expand_env_vars


class UndefinedEnvVarError(Error):
    """Occurs when an environment variable is used as a command argument but is not set."""

    def __init__(self, command_name, var):
        """Initialize UndefinedEnvVarError."""
        super().__init__(
            f"Command '{command_name}' referenced unset environment variable '{var}' in an argument. "
            + "Set the environment variable or update the command definition."
        )


class Command(Canonical):
    """This class represents stored command arguments for plugins."""

    def __init__(self, args: str, description: Optional[str] = None):
        """Initialize a Command."""
        super().__init__(args=args, description=description)

    def expanded_args(self, name, env):
        """
        Replace any env var arguments with their values.

        :raises UndefinedEnvVarError: if an env var argument is not set
        """
        expanded = []
        for arg in shlex.split(self.args):
            value = expand_env_vars(arg, env)
            if not value:
                raise UndefinedEnvVarError(name, arg)

            expanded.append(value)

        return expanded

    def canonical(self):
        """Serialize the command."""
        return Command.as_canonical(self)

    @classmethod
    def as_canonical(cls, target):
        """Serialize the target command."""
        canonical = super().as_canonical(target)
        # if there are only args, flatten the object
        # to the short form
        if "args" in canonical and len(canonical) == 1:
            return canonical["args"]

        return canonical

    @classmethod
    def parse(cls, obj):
        """Deserialize data into a Command."""
        if isinstance(obj, str):
            # allow setting the arguments as the value
            # without a description
            return Command(args=obj)

        return super().parse(obj)

    @classmethod
    def parse_all(cls, obj):
        """Deserialize commands data into a dict of Commands."""
        if isinstance(obj, dict):
            return {name: Command.parse(cmd) for name, cmd in obj.items()}

        if obj is None:
            return {}

        raise ValueError(f"Expected command to be a dict but was {type(obj)}")



from pypika import Field
from pypika import functions as fn
from pypika.terms import Function


class DateTrunc(Function):
    def __init__(self, part, term, alias=None):
        super().__init__("DATE_TRUNC", part, term, alias=alias)


class Date:
    def __init__(self, timeframe, table, name):
        self.sql = None
        field = Field(name, table=table)
        if timeframe == "date*":
            alias = "{}_date".format(field.get_sql().replace("_", ""))
            self.sql = fn.Date(field, alias=alias)
        elif timeframe == "month*":
            alias = "{}_month".format(field.get_sql().replace("_", ""))
            self.sql = fn.ToChar(DateTrunc("month", field), "YYYY-MM", alias=alias)
        elif timeframe == "week*":
            alias = "{}_week".format(field.get_sql().replace("_", ""))
            self.sql = fn.ToChar(DateTrunc("week", field), "YYYY-MM-DD", alias=alias)
        elif timeframe == "year*":
            alias = "{}_year".format(field.get_sql().replace("_", ""))
            self.sql = fn.Extract("YEAR", field, alias=alias)



"""Base Error classes."""
import functools
import io
import logging
import subprocess
from enum import Enum
from typing import Optional, Union


class ExitCode(int, Enum):
    OK = 0
    FAIL = 1
    NO_RETRY = 2


class Error(Exception):
    """Base exception for ELT errors."""

    def exit_code(self):
        return ExitCode.FAIL


class ExtractError(Error):
    """Error in the extraction process, like API errors."""

    def exit_code(self):
        return ExitCode.NO_RETRY


class SubprocessError(Exception):
    """Happens when subprocess exits with a resultcode != 0"""

    def __init__(
        self,
        message: str,
        process: subprocess.CompletedProcess,
        stderr: Optional[Union[str, bytes, io.TextIOBase]] = None,
    ):
        """Initialize SubprocessError."""
        self.process = process
        self._stderr = stderr or process.stderr
        super().__init__(message)

    @property
    def stderr(self) -> Optional[str]:
        """Return the output of the process to stderr."""
        if not self._stderr:
            return None
        elif isinstance(self._stderr, bytes):
            self._stderr = self._stderr.decode("utf-8")
        elif not isinstance(self._stderr, str):
            self._stderr = self._stderr.read()

        return self._stderr


class PluginInstallError(Exception):
    """Happens when a plugin fails to install."""

    pass


class PluginInstallWarning(Exception):
    """Happens when a plugin optional optional step fails to install."""

    pass


class PluginNotInstallable(Exception):
    pass


def aggregate(error_cls):
    class Aggregate(error_cls):
        """Aggregate multiple sub-exceptions."""

        def __init__(self, exceptions: []):
            self.exceptions = exceptions

        def __str__(self):
            return "\n".join(str(e) for e in self.exceptions)

    if error_cls != Exception:
        error_cls.Aggregate = Aggregate

    return error_cls


AggregateError = aggregate(Exception)


@aggregate
class ImportError(Error):
    """
    Error in the import process, the data cannot be processed.
    """

    def __init__(self, entry):
        self.entry = entry


@aggregate
class SchemaError(Error):
    """Base exception for schema errors."""


@aggregate
class InapplicableChangeError(SchemaError):
    """Raise for inapplicable schema changes."""

    def exit_code(self):
        return ExitCode.NO_RETRY


# TODO: use as a context manager instead
class ExceptionAggregator:
    def __init__(self, etype):
        self.success = []
        self.failures = []
        self.etype = etype

    def recognize_exception(self, e) -> bool:
        return self.etype == type(e)

    def call(self, callable, *args, **kwargs):
        params = (args, kwargs)
        try:
            ret = callable(*args, **kwargs)
            self.success.append(params)
            return ret
        except Exception as e:
            if self.recognize_exception(e):
                self.failures.append((e, params))
            else:
                raise e

    def raise_aggregate(self):
        aggregate_type = AggregateError

        if hasattr(self.etype, "Aggregate"):
            aggregate_type = self.etype.Aggregate

        if len(self.failures):
            exceptions = [f[0] for f in self.failures]
            raise aggregate_type(exceptions)


def with_error_exit_code(main):
    @functools.wraps(main)
    def f(*args, **kwargs):
        try:
            main(*args, **kwargs)
        except Error as err:
            logging.error(str(err))
            exit(err.exit_code())
        except Exception as e:
            logging.error(str(e))
            raise e

    return f



"""Defines StaleJobFailer."""
import logging

from .finder import JobFinder

logger = logging.getLogger(__name__)


class StaleJobFailer:
    """Class that will look for stale jobs and mark them as failed."""

    def __init__(self, job_id=None):
        """Initialize stale job failer with optional job ID to filter by."""
        self.job_id = job_id

    def fail_stale_jobs(self, session):
        """Mark all stale jobs as failed."""
        for job in self._stale_jobs(session):
            self._fail_stale_job(job, session)

    def _stale_jobs(self, session):
        if self.job_id:
            return JobFinder(self.job_id).stale(session)

        return JobFinder.all_stale(session)

    def _fail_stale_job(self, job, session):
        if not job.fail_stale():
            return

        job.save(session)

        # No need to mention job ID if they're all going to be the same.
        with_job_id = "" if self.job_id else f" with job ID '{job.job_id}'"

        error = job.payload["error"]
        logger.info(
            f"Marked stale run{with_job_id} that started at {job.started_at} as failed: {error}"
        )



"""Defines helpers for use by the CLI."""
import logging
import os
from typing import Optional

import click
from meltano.core.logging import setup_logging
from meltano.core.plugin import PluginType
from meltano.core.plugin_install_service import (
    PluginInstallReason,
    PluginInstallService,
    PluginInstallStatus,
)
from meltano.core.project import Project
from meltano.core.project_add_service import (
    PluginAlreadyAddedException,
    ProjectAddService,
)
from meltano.core.setting_definition import SettingKind
from meltano.core.tracking import GoogleAnalyticsTracker

setup_logging()

logger = logging.getLogger(__name__)


class CliError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.printed = False

    def print(self):
        if self.printed:
            return

        logger.debug(str(self), exc_info=True)
        click.secho(str(self), fg="red")

        self.printed = True


def print_added_plugin(project, plugin, related=False):
    descriptor = plugin.type.descriptor
    if related:
        descriptor = f"related {descriptor}"

    if plugin.should_add_to_file():
        click.secho(
            f"Added {descriptor} '{plugin.name}' to your Meltano project", fg="green"
        )
    else:
        click.secho(
            f"Adding {descriptor} '{plugin.name}' to your Meltano project...",
            fg="green",
        )

    inherit_from = plugin.inherit_from
    has_variant = plugin.is_variant_set
    variant_label = plugin.variant_label(plugin.variant)
    if inherit_from:
        if has_variant:
            inherit_from = f"{inherit_from}, variant {variant_label}"

        click.echo(f"Inherit from:\t{inherit_from}")
    elif has_variant:
        click.echo(f"Variant:\t{variant_label}")

    repo_url = plugin.repo
    if repo_url:
        click.echo(f"Repository:\t{repo_url}")

    docs_url = plugin.docs
    if docs_url:
        click.echo(f"Documentation:\t{docs_url}")


def _prompt_plugin_namespace(plugin_type, plugin_name):
    click.secho(
        f"Adding new custom {plugin_type.descriptor} with name '{plugin_name}'...",
        fg="green",
    )
    click.echo()

    click.echo(
        f"Specify the plugin's {click.style('namespace', fg='blue')}, which will serve as the:"
    )
    click.echo("- identifier to find related/compatible plugins")
    if plugin_type == PluginType.EXTRACTORS:
        click.echo("- default database schema (`load_schema` extra),")
        click.echo("  for use by loaders that support a target schema")
    elif plugin_type == PluginType.LOADERS:
        click.echo("- default target database dialect (`dialect` extra),")
        click.echo("  for use by transformers that connect with the database")
    click.echo()

    click.echo(
        "Hit Return to accept the default: plugin name with underscores instead of dashes"
    )
    click.echo()

    return click.prompt(
        click.style("(namespace)", fg="blue"),
        type=str,
        default=plugin_name.replace("-", "_"),
    )


def _prompt_plugin_pip_url(plugin_name: str) -> Optional[str]:
    click.echo()
    click.echo(
        f"Specify the plugin's {click.style('`pip install` argument', fg='blue')}, for example:"
    )
    click.echo("- PyPI package name:")
    click.echo(f"\t{plugin_name}")
    click.echo("- Git repository URL:")
    click.echo(f"\tgit+https://gitlab.com/meltano/{plugin_name}.git")
    click.echo("- local directory, in editable/development mode:")
    click.echo(f"\t-e extract/{plugin_name}")
    click.echo("- 'n' if using a local executable (nothing to install)")
    click.echo()
    click.echo("Default: plugin name as PyPI package name")
    click.echo()

    result = click.prompt(
        click.style("(pip_url)", fg="blue"), type=str, default=plugin_name
    )
    return None if result == "n" else result


def _prompt_plugin_executable(pip_url: Optional[str], plugin_name: str) -> str:
    derived_from = "`pip_url`"
    prompt_request = "executable name"
    if pip_url is None:
        derived_from = "the plugin name"
        prompt_request = "executable path"

    click.echo()
    click.echo(f"Specify the plugin's {click.style(prompt_request, fg='blue')}")
    click.echo()
    click.echo(f"Default: name derived from {derived_from}")
    click.echo()

    plugin_basename = os.path.basename(pip_url or plugin_name)
    package_name, _ = os.path.splitext(plugin_basename)
    return click.prompt(click.style("(executable)", fg="blue"), default=package_name)


def _prompt_plugin_capabilities(plugin_type):
    if plugin_type != PluginType.EXTRACTORS:
        return []

    click.echo()
    click.echo(
        f"Specify the tap's {click.style('supported Singer features', fg='blue')} (executable flags), for example:"
    )
    click.echo("\t`catalog`: supports the `--catalog` flag")
    click.echo("\t`discover`: supports the `--discover` flag")
    click.echo("\t`properties`: supports the `--properties` flag")
    click.echo("\t`state`: supports the `--state` flag")
    click.echo()
    click.echo(
        "To find out what features a tap supports, reference its documentation or try one"
    )
    click.echo(
        "of the tricks under https://meltano.com/docs/contributor-guide.html#how-to-test-a-tap."
    )
    click.echo()
    click.echo("Multiple capabilities can be separated using commas.")
    click.echo()
    click.echo("Default: no capabilities")
    click.echo()

    return click.prompt(
        click.style("(capabilities)", fg="blue"),
        type=list,
        default=[],
        value_proc=lambda value: [word.strip() for word in value.split(",")],
    )


def _prompt_plugin_settings(plugin_type):
    if plugin_type not in {PluginType.EXTRACTORS, PluginType.LOADERS}:
        return []
    singer_type = "tap" if plugin_type == PluginType.EXTRACTORS else "target"

    click.echo()
    click.echo(
        f"Specify the {singer_type}'s {click.style('supported settings', fg='blue')} (`config.json` keys)"
    )
    click.echo()
    click.echo("Multiple setting names (keys) can be separated using commas.")
    click.echo()
    click.echo(
        "A setting kind can be specified alongside the name (key) by using the `:` delimiter,"
    )
    click.echo("e.g. `port:integer` to set the kind `integer` for the name `port`")
    click.echo()
    click.echo("Supported setting kinds:")
    click.echo(
        " | ".join([click.style(kind.value, fg="magenta") for kind in SettingKind])
    )
    click.echo()
    click.echo(
        "- Credentials and other sensitive setting types should use the "
        + click.style("password", fg="magenta")
        + " kind."
    )
    click.echo(
        "- If not specified, setting kind defaults to "
        + click.style("string", fg="magenta")
        + "."
    )
    click.echo(
        "- Nested properties can be represented using the `.` separator, "
        + 'e.g. `auth.username` for `{ "auth": { "username": value } }`.'
    )
    click.echo(
        f"- To find out what settings a {singer_type} supports, reference its documentation."
    )
    click.echo()
    click.echo("Default: no settings")
    click.echo()

    settings: dict = None
    while settings is None:  # noqa:  WPS426  # allows lambda in loop
        settings_input = click.prompt(
            click.style("(settings)", fg="blue"),
            type=list,
            default=[],
            value_proc=lambda value: [
                setting.strip().partition(":") for setting in value.split(",")
            ],
        )
        try:
            settings = [
                {"name": name, "kind": kind and SettingKind(kind).value}
                for name, sep, kind in settings_input
            ]
        except ValueError as ex:
            click.secho(str(ex), fg="red")

    return settings


def add_plugin(
    project: Project,
    plugin_type: PluginType,
    plugin_name: str,
    add_service: ProjectAddService,
    variant=None,
    inherit_from=None,
    custom=False,
):
    plugin_attrs = {}
    if custom:
        namespace = _prompt_plugin_namespace(plugin_type, plugin_name)
        pip_url = _prompt_plugin_pip_url(plugin_name)
        executable = _prompt_plugin_executable(pip_url, plugin_name)
        capabilities = _prompt_plugin_capabilities(plugin_type)
        settings = _prompt_plugin_settings(plugin_type)

        plugin_attrs = {
            "namespace": namespace,
            "pip_url": pip_url,
            "executable": executable,
            "capabilities": capabilities,
            "settings": settings,
        }

    try:
        plugin = add_service.add(
            plugin_type,
            plugin_name,
            variant=variant,
            inherit_from=inherit_from,
            **plugin_attrs,
        )
        print_added_plugin(project, plugin)
    except PluginAlreadyAddedException as err:
        plugin = err.plugin

        click.secho(
            f"{plugin_type.descriptor.capitalize()} '{plugin_name}' already exists in your Meltano project",
            fg="yellow",
            err=True,
        )

        if variant and variant != plugin.variant:
            variant = plugin.find_variant(variant)

            click.echo()
            click.echo(
                f"To switch from the current '{plugin.variant}' variant to '{variant.name}':"
            )
            click.echo(
                "1. Update `variant` and `pip_url` in your `meltano.yml` project file:"
            )
            click.echo(f"\tname: {plugin.name}")
            click.echo(f"\tvariant: {variant.name}")
            click.echo(f"\tpip_url: {variant.pip_url}")

            click.echo("2. Reinstall the plugin:")
            click.echo(f"\tmeltano install {plugin_type.singular} {plugin.name}")

            click.echo(
                "3. Check if the configuration is still valid (and make changes until it is):"
            )
            click.echo(f"\tmeltano config {plugin.name} list")
            click.echo()
            click.echo(
                "To learn more, visit https://www.meltano.com/docs/plugin-management.html#switching-from-one-variant-to-another"
            )

            click.echo()
            click.echo(
                f"Alternatively, to keep the existing '{plugin.name}' with variant '{plugin.variant}',"
            )
            click.echo(
                f"add variant '{variant.name}' as a separate plugin with its own unique name:"
            )
            click.echo(
                "\tmeltano add {type} {name}--{variant} --inherit-from {name} --variant {variant}".format(
                    type=plugin_type.singular, name=plugin.name, variant=variant.name
                )
            )

            click.echo()
            click.echo(
                "To learn more, visit https://www.meltano.com/docs/plugin-management.html#multiple-variants"
            )
        else:
            click.echo(
                "To add it to your project another time so that each can be configured differently,"
            )
            click.echo(
                "add a new plugin inheriting from the existing one with its own unique name:"
            )
            click.echo(
                f"\tmeltano add {plugin_type.singular} {plugin.name}--new --inherit-from {plugin.name}"
            )

    click.echo()

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type=plugin_type, plugin_name=plugin_name)

    return plugin


def add_related_plugins(
    project, plugins, add_service: ProjectAddService, plugin_types=list(PluginType)
):
    added_plugins = []
    for plugin_install in plugins:
        related_plugins = add_service.add_related(
            plugin_install, plugin_types=plugin_types
        )
        for related_plugin in related_plugins:
            print_added_plugin(project, related_plugin, related=True)
            click.echo()

        added_plugins.extend(related_plugins)

    return added_plugins


def install_status_update(install_state):
    """
    Print the status of plugin installation.

    Used as the callback for PluginInstallService.
    """
    plugin = install_state.plugin
    desc = plugin.type.descriptor
    if install_state.status is PluginInstallStatus.RUNNING:
        msg = f"{install_state.verb} {desc} '{plugin.name}'..."
        click.secho(msg)
    elif install_state.status is PluginInstallStatus.ERROR:
        click.secho(install_state.message, fg="red")
        click.secho(install_state.details, err=True)
    elif install_state.status is PluginInstallStatus.WARNING:
        click.secho(f"Warning! {install_state.message}.", fg="yellow")
    elif install_state.status is PluginInstallStatus.SUCCESS:
        msg = f"{install_state.verb} {desc} '{plugin.name}'"
        click.secho(msg, fg="green")
        click.echo()


def install_plugins(
    project, plugins, reason=PluginInstallReason.INSTALL, parallelism=None
):
    """Install the provided plugins and report results to the console."""
    install_service = PluginInstallService(
        project, status_cb=install_status_update, parallelism=parallelism
    )
    install_results = install_service.install_plugins(plugins, reason=reason)
    num_successful = len([status for status in install_results if status.successful])
    num_skipped = len([status for status in install_results if status.skipped])
    num_failed = len(install_results) - num_successful

    fg = "green"
    if num_failed >= 0 and num_successful == 0:
        fg = "red"
    elif num_failed > 0 and num_successful > 0:
        fg = "yellow"

    if len(plugins) > 1:
        verb = "Updated" if reason == PluginInstallReason.UPGRADE else "Installed"
        click.secho(
            f"{verb} {num_successful-num_skipped}/{num_successful+num_failed} plugins",
            fg=fg,
        )
    if num_skipped:
        verb = "Skipped installing"
        click.secho(
            f"{verb} {num_skipped}/{num_successful+num_failed} plugins",
            fg=fg,
        )

    return num_failed == 0



"""Defines `meltano remove` command."""
import click
from meltano.core.plugin import PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_location_remove import (
    DbRemoveManager,
    PluginLocationRemoveManager,
)
from meltano.core.plugin_remove_service import PluginRemoveService

from . import cli
from .params import pass_project


@cli.command()
@click.argument("plugin_type", type=click.Choice(PluginType.cli_arguments()))
@click.argument("plugin_names", nargs=-1, required=True)
@pass_project()
@click.pass_context
def remove(ctx, project, plugin_type, plugin_names):
    """Remove a plugin from your project."""
    plugins = [
        ProjectPlugin(PluginType.from_cli_argument(plugin_type), plugin_name)
        for plugin_name in plugin_names
    ]

    remove_plugins(project, plugins)


def remove_plugins(project, plugins):
    """Invoke PluginRemoveService and output CLI removal overview."""
    remove_service = PluginRemoveService(project)

    num_removed, total = remove_service.remove_plugins(
        plugins,
        plugin_status_cb=remove_plugin_status_update,
        removal_manager_status_cb=removal_manager_status_update,
    )

    click.echo()
    fg = "green"
    if num_removed < total:
        fg = "yellow"

    if len(plugins) > 1:
        click.secho(f"Fully removed {num_removed}/{total} plugins", fg=fg)
        click.echo()


def remove_plugin_status_update(plugin):
    """Print remove status message for a plugin."""
    plugin_descriptor = f"{plugin.type.descriptor} '{plugin.name}'"

    click.echo()
    click.secho(f"Removing {plugin_descriptor}...")
    click.echo()


def removal_manager_status_update(removal_manager: PluginLocationRemoveManager):
    """Print remove status message for a plugin location."""
    plugin_descriptor = removal_manager.plugin_descriptor
    location = removal_manager.location
    message = removal_manager.remove_message

    if removal_manager.plugin_error:
        click.secho(
            f"Error removing plugin {plugin_descriptor} from {location}: {message}",
            fg="red",
        )

    elif removal_manager.plugin_not_found:
        click.secho(
            f"Could not find {plugin_descriptor} in {location} to remove", fg="yellow"
        )

    elif removal_manager.plugin_removed:

        msg = f"Removed {plugin_descriptor} from {location}"

        if isinstance(removal_manager, DbRemoveManager):
            msg = f"Reset {plugin_descriptor} plugin settings in the {location}"

        click.secho(msg, fg="green")



import itertools
from contextlib import contextmanager

import click
from meltano.core.compiler.project_compiler import ProjectCompiler

from . import cli
from .params import pass_project


def indent(text: str, count=2, char="  "):
    return "".join(list(itertools.repeat(char, count))) + text


@cli.group(hidden=True, invoke_without_command=True)
@pass_project()
@click.pass_context
def model(ctx, project):
    compiler = ctx.obj["compiler"] = ProjectCompiler(project)
    compiler.parse()

    if ctx.invoked_subcommand is None:
        click.secho("Local", fg="green")
        for model in compiler.topics:
            click.echo(f'{model["namespace"]}/{model["name"]}')

        print()

        click.secho("Packaged", fg="green")
        for model in compiler.package_topics:
            click.echo(f'{model["namespace"]}/{model["name"]}')


@model.command()
@click.pass_context
def compile(ctx):
    compiler = ctx.obj["compiler"]

    compiler.compile()
    click.secho(f"Compiled {len(compiler.topics + compiler.package_topics)} model(s)")


@model.command()
@click.pass_context
def show(ctx):
    compiler = ctx.obj["compiler"]

    def print_table(table: dict):
        print(indent(table["name"], count=2))

        for dim in table.get("columns", []):
            print(indent(f"[C] {dim['name']} ({dim['sql']})", count=3))

        for dim in table.get("timeframes", []):
            print(indent(f"[T] {dim['name']} ({dim['sql']})", count=3))

        for dim in table.get("aggregates", []):
            print(indent(f"[A] {dim['name']} ({dim['sql']})", count=3))

    for model in compiler.package_topics:
        print(model["namespace"])

        for design in model["designs"]:
            print(indent(design["name"], count=1))
            print_table(design["related_table"])

            for join in design.get("joins", []):
                print_table(join["related_table"])



import json
import uuid

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import CHAR, INTEGER, VARCHAR, TypeDecorator


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::
        JSONEncodedDict(255)
    """

    impl = VARCHAR
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class IntFlag(TypeDecorator):
    impl = INTEGER
    cache_ok = True

    # force the cast to INTEGER
    def process_bind_param(self, value, dialect):
        return int(value)


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    Reference: https://docs.sqlalchemy.org/en/13/core/custom_types.html#backend-agnostic-guid-type
    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value



import logging
import os
import shutil
from functools import singledispatch
from pathlib import Path
from typing import Dict, List

import click

from .db import project_engine
from .migration_service import MigrationError, MigrationService
from .plugin.meltano_file import MeltanoFilePlugin
from .project import Project
from .project_settings_service import ProjectSettingsService, SettingValueStore
from .utils import truthy


class ProjectInitServiceError(Exception):
    pass


class ProjectInitService:
    def __init__(self, project_name):
        self.project_name = project_name.lower()

    def init(self, activate=True, add_discovery=False) -> Project:
        try:
            os.mkdir(self.project_name)
        except Exception as e:
            raise ProjectInitServiceError(
                f"Directory {self.project_name} already exists!"
            )
        click.secho(f"Created", fg="blue", nl=False)
        click.echo(f" {self.project_name}")

        self.project = Project(self.project_name)
        self.settings_service = ProjectSettingsService(self.project)

        self.create_files(add_discovery=add_discovery)
        self.set_send_anonymous_usage_stats()

        if activate:
            Project.activate(self.project)

        self.create_system_database()

        return self.project

    def create_files(self, add_discovery=False):
        click.secho(f"Creating project files...", fg="blue")

        plugin = MeltanoFilePlugin(discovery=add_discovery)
        for path in plugin.create_files(self.project):
            click.secho(f"Created", fg="blue", nl=False)
            click.echo(f" {self.project_name}/{path}")

    def set_send_anonymous_usage_stats(self):
        # Ensure the value is explicitly stored in `meltano.yml`
        self.settings_service.set(
            "send_anonymous_usage_stats",
            self.settings_service.get("send_anonymous_usage_stats"),
            store=SettingValueStore.MELTANO_YML,
        )

    def create_system_database(self):
        click.secho(f"Creating system database...", fg="blue")

        # register the system database connection
        engine, _ = project_engine(self.project, default=True)

        try:
            migration_service = MigrationService(engine)
            migration_service.upgrade()
            migration_service.seed(self.project)
        except MigrationError as err:
            raise ProjectInitServiceError(str(err)) from err

    def echo_instructions(self):
        click.echo()
        click.secho("Project", nl=False)
        click.secho(f" {self.project_name}", fg="green", nl=False)
        click.echo(" has been created.")

        click.echo("\nNext steps:")
        click.secho("\tcd ", nl=False)
        click.secho(self.project_name, fg="green")
        click.echo("\tVisit https://meltano.com/ to learn where to go from here")

        click.echo()
        click.secho("> ", fg="bright_black", nl=False)

        click.secho(
            "Meltano sends anonymous usage data that helps improve the product."
        )

        click.secho("> ", fg="bright_black", nl=False)

        click.echo("You can opt-out for new, existing, or all projects.")

        click.secho("> ", fg="bright_black", nl=False)

        click.secho(
            "https://meltano.com/docs/settings.html#send-anonymous-usage-stats",
            fg="cyan",
        )

    def join_with_project_base(self, filename):
        return os.path.join(".", self.project_name, filename)



import asyncio
import json
import logging
import os
import shutil
import subprocess
import sys
from contextlib import suppress
from datetime import datetime
from enum import IntFlag
from pathlib import Path

from meltano.core.elt_context import ELTContext
from meltano.core.job import Job, Payload
from meltano.core.logging import capture_subprocess_output
from meltano.core.plugin import PluginType
from meltano.core.plugin.singer import PluginType, SingerTap, SingerTarget
from meltano.core.plugin_invoker import PluginInvoker, invoker_factory
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import human_size

from . import Runner, RunnerError


class BookmarkWriter:
    def __init__(self, job, session, payload_flag=Payload.STATE):
        self.job = job
        self.session = session
        self.payload_flag = payload_flag

    def writeline(self, line):
        if self.job is None:
            logging.info(
                f"Running outside a Job context: incremental state could not be updated."
            )
            return

        try:
            new_state = json.loads(line)
            job = self.job

            job.payload["singer_state"] = new_state
            job.payload_flags |= self.payload_flag
            job.save(self.session)

            logging.info(f"Incremental state has been updated at {datetime.utcnow()}.")
            logging.debug(f"Incremental state: {new_state}")
        except Exception as err:
            logging.warning(
                "Received state is invalid, incremental state has not been updated"
            )


class SingerRunner(Runner):
    def __init__(self, elt_context: ELTContext):
        self.context = elt_context

        self.project_settings_service = ProjectSettingsService(
            self.context.project,
            config_service=self.context.plugins_service.config_service,
        )

    def stop(self, process, **wait_args):
        while True:
            try:
                code = process.wait(**wait_args)
                logging.debug(f"{process} exited with {code}")
                return code
            except subprocess.TimeoutExpired:
                process.kill()
                logging.error(f"{process} was killed.")

    async def invoke(  # noqa: WPS217, WPS210, WPS213, WPS231
        self,
        tap: PluginInvoker,
        target: PluginInvoker,
        extractor_log=None,
        loader_log=None,
        extractor_out=None,
        loader_out=None,
    ):
        """Invoke tap and target together."""
        extractor_log = extractor_log or sys.stderr
        loader_log = loader_log or sys.stderr

        # The StreamReader line length limit also acts as half the buffer size,
        # which cannot be set directly:
        # - https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L395-L396
        # - https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L482
        stream_buffer_size = self.project_settings_service.get("elt.buffer_size")
        line_length_limit = stream_buffer_size // 2

        # Start tap
        try:
            p_tap = await tap.invoke_async(
                limit=line_length_limit,
                stdout=asyncio.subprocess.PIPE,  # Singer messages
                stderr=asyncio.subprocess.PIPE,  # Log
            )
        except Exception as err:
            raise RunnerError(f"Cannot start extractor: {err}") from err

        # Start target
        try:
            p_target = await target.invoke_async(
                limit=line_length_limit,
                stdin=asyncio.subprocess.PIPE,  # Singer messages
                stdout=asyncio.subprocess.PIPE,  # Singer state
                stderr=asyncio.subprocess.PIPE,  # Log
            )
        except Exception as err:
            raise RunnerError(f"Cannot start loader: {err}") from err

        # Process tap output
        tap_outputs = [p_target.stdin]
        if extractor_out:
            tap_outputs.insert(0, extractor_out)

        tap_stdout_future = asyncio.ensure_future(
            capture_subprocess_output(p_tap.stdout, *tap_outputs)
        )
        tap_stderr_future = asyncio.ensure_future(
            capture_subprocess_output(p_tap.stderr, extractor_log)
        )

        # Process target output
        target_outputs = [self.bookmark_writer()]
        if loader_out:
            target_outputs.insert(0, loader_out)

        target_stdout_future = asyncio.ensure_future(
            capture_subprocess_output(p_target.stdout, *target_outputs)
        )
        target_stderr_future = asyncio.ensure_future(
            capture_subprocess_output(p_target.stderr, loader_log)
        )

        # Wait for tap or target to complete, or for one of the output handlers to raise an exception.
        tap_process_future = asyncio.ensure_future(p_tap.wait())
        target_process_future = asyncio.ensure_future(p_target.wait())
        output_exception_future = asyncio.ensure_future(
            asyncio.wait(
                [
                    tap_stdout_future,
                    tap_stderr_future,
                    target_stdout_future,
                    target_stderr_future,
                ],
                return_when=asyncio.FIRST_EXCEPTION,
            ),
        )

        done, _ = await asyncio.wait(
            [tap_process_future, target_process_future, output_exception_future],
            return_when=asyncio.FIRST_COMPLETED,
        )

        # If `output_exception_future` completes first, one of the output handlers raised an exception or all completed successfully.
        if output_exception_future in done:
            output_futures_done, _ = output_exception_future.result()
            output_futures_failed = [
                future
                for future in output_futures_done
                if future.exception() is not None
            ]

            if output_futures_failed:
                # If any output handler raised an exception, re-raise it.

                # Special behavior for the tap stdout handler raising a line length limit error.
                if tap_stdout_future in output_futures_failed:
                    self._handle_tap_line_length_limit_error(
                        tap_stdout_future.exception(),
                        line_length_limit=line_length_limit,
                        stream_buffer_size=stream_buffer_size,
                    )

                failed_future = output_futures_failed.pop()
                raise failed_future.exception()
            else:
                # If all of the output handlers completed without raising an exception,
                # we still need to wait for the tap or target to complete.
                done, _ = await asyncio.wait(
                    [tap_process_future, target_process_future],
                    return_when=asyncio.FIRST_COMPLETED,
                )

        if target_process_future in done:
            target_code = target_process_future.result()

            if tap_process_future in done:
                tap_code = tap_process_future.result()
            else:
                # If the target completes before the tap, it failed before processing all tap output

                # Kill tap and cancel output processing since there's no more target to forward messages to
                p_tap.kill()
                await tap_process_future
                tap_stdout_future.cancel()
                tap_stderr_future.cancel()

                # Pretend the tap finished successfully since it didn't itself fail
                tap_code = 0

            # Wait for all buffered target output to be processed
            await asyncio.wait([target_stdout_future, target_stderr_future])
        else:  # if tap_process_future in done:
            # If the tap completes before the target, the target should have a chance to process all tap output
            tap_code = tap_process_future.result()

            # Wait for all buffered tap output to be processed
            await asyncio.wait([tap_stdout_future, tap_stderr_future])

            # Close target stdin so process can complete naturally
            p_target.stdin.close()
            with suppress(AttributeError):  # `wait_closed` is Python 3.7+
                await p_target.stdin.wait_closed()

            # Wait for all buffered target output to be processed
            await asyncio.wait([target_stdout_future, target_stderr_future])

            # Wait for target to complete
            target_code = await target_process_future

        if tap_code and target_code:
            raise RunnerError(
                "Extractor and loader failed",
                {PluginType.EXTRACTORS: tap_code, PluginType.LOADERS: target_code},
            )
        elif tap_code:
            raise RunnerError("Extractor failed", {PluginType.EXTRACTORS: tap_code})
        elif target_code:
            raise RunnerError("Loader failed", {PluginType.LOADERS: target_code})

    def bookmark_writer(self):
        incomplete_state = self.context.full_refresh and self.context.select_filter
        payload_flag = Payload.INCOMPLETE_STATE if incomplete_state else Payload.STATE
        return BookmarkWriter(
            self.context.job, self.context.session, payload_flag=payload_flag
        )

    def dry_run(self, tap: PluginInvoker, target: PluginInvoker):
        logging.info("Dry run:")
        logging.info(f"\textractor: {tap.plugin.name} at '{tap.exec_path()}'")
        logging.info(f"\tloader: {target.plugin.name} at '{target.exec_path()}'")

    async def run(
        self, extractor_log=None, loader_log=None, extractor_out=None, loader_out=None
    ):
        tap = self.context.extractor_invoker()
        target = self.context.loader_invoker()

        if self.context.dry_run:
            return self.dry_run(tap, target)

        with tap.prepared(self.context.session), target.prepared(self.context.session):
            await self.invoke(
                tap,
                target,
                extractor_log=extractor_log,
                loader_log=loader_log,
                extractor_out=extractor_out,
                loader_out=loader_out,
            )

    def _handle_tap_line_length_limit_error(
        self,
        exception,
        line_length_limit,
        stream_buffer_size,
    ):
        # StreamReader.readline can raise a ValueError wrapping a LimitOverrunError:
        # https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L549
        if not isinstance(exception, ValueError):
            return

        exception = exception.__context__  # noqa: WPS609
        if not isinstance(exception, asyncio.LimitOverrunError):
            return

        logging.error(
            f"The extractor generated a message exceeding the message size limit of {human_size(line_length_limit)} (half the buffer size of {human_size(stream_buffer_size)})."
        )
        logging.error(
            "To let this message be processed, increase the 'elt.buffer_size' setting to at least double the size of the largest expected message, and try again."
        )
        logging.error(
            "To learn more, visit https://www.meltano.com/docs/settings.html#elt-buffer-size"
        )
        raise RunnerError("Output line length limit exceeded") from exception



import datetime

from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical
from meltano.core.job import JobFinder

CRON_INTERVALS = {
    "@once": None,
    "@hourly": "0 * * * *",
    "@daily": "0 0 * * *",
    "@weekly": "0 0 * * 0",
    "@monthly": "0 0 1 * *",
    "@yearly": "0 0 1 1 *",
}


class Schedule(NameEq, Canonical):
    def __init__(
        self,
        name: str = None,
        extractor: str = None,
        loader: str = None,
        transform: str = None,
        interval: str = None,
        start_date: datetime = None,
        env={},
    ):
        super().__init__()

        # Attributes will be listed in meltano.yml in this order:
        self.name = name
        self.extractor = extractor
        self.loader = loader
        self.transform = transform
        self.interval = interval
        self.start_date = start_date
        self.env = env

    @property
    def cron_interval(self):
        return CRON_INTERVALS.get(self.interval, self.interval)

    @property
    def elt_args(self):
        return [
            self.extractor,
            self.loader,
            f"--transform={self.transform}",
            f"--job_id={self.name}",
        ]

    def last_successful_run(self, session):
        return JobFinder(self.name).latest_success(session)



import asyncio
import copy
import logging
import os
import subprocess
from contextlib import contextmanager
from typing import Optional

from .error import Error, SubprocessError
from .plugin import PluginRef
from .plugin.config_service import PluginConfigService
from .plugin.project_plugin import ProjectPlugin
from .plugin.settings_service import PluginSettingsService
from .project import Project
from .project_plugins_service import ProjectPluginsService
from .venv_service import VenvService, VirtualEnv


def invoker_factory(project, plugin: ProjectPlugin, *args, **kwargs):
    cls = PluginInvoker

    if hasattr(plugin, "invoker_class"):
        cls = plugin.invoker_class

    invoker = cls(project, plugin, *args, **kwargs)

    return invoker


class InvokerError(Error):
    pass


class ExecutableNotFoundError(InvokerError):
    """Occurs when the executable could not be found"""

    def __init__(self, plugin: PluginRef, executable):
        super().__init__(
            f"Executable '{executable}' could not be found. "
            f"{plugin.type.descriptor.capitalize()} '{plugin.name}' may not have been installed yet using `meltano install {plugin.type.singular} {plugin.name}`, or the executable name may be incorrect."
        )


class InvokerNotPreparedError(InvokerError):
    """Occurs when `invoke` is called before `prepare`"""

    pass


class UnknownCommandError(InvokerError):
    """Occurs when `invoke` is called in command mode with an undefined command."""

    def __init__(self, plugin: PluginRef, command):
        """Initialize UnknownCommandError."""
        self.plugin = plugin
        self.command = command

    def __str__(self):
        """Return error message."""
        if self.plugin.supported_commands:
            supported_commands = ", ".join(self.plugin.supported_commands)
            desc = f"supports the following commands: {supported_commands}"
        else:
            desc = "does not define any commands."
        return " ".join(
            [
                f"Command '{self.command}' could not be found.",
                f"{self.plugin.type.descriptor.capitalize()} '{self.plugin.name}'",
                desc,
            ]
        )


class PluginInvoker:
    """This class handles the invocation of a `ProjectPlugin` instance."""

    def __init__(
        self,
        project: Project,
        plugin: ProjectPlugin,
        context: Optional[object] = None,
        run_dir=None,
        config_dir=None,
        venv_service: VenvService = None,
        plugins_service: ProjectPluginsService = None,
        plugin_config_service: PluginConfigService = None,
        plugin_settings_service: PluginSettingsService = None,
    ):
        self.project = project
        self.plugin = plugin
        self.context = context

        self.venv_service: Optional[VenvService] = None
        if plugin.pip_url or venv_service:
            self.venv_service = venv_service or VenvService(
                project,
                name=plugin.name,
                namespace=plugin.type,
            )
        self.plugin_config_service = plugin_config_service or PluginConfigService(
            plugin,
            config_dir or self.project.plugin_dir(plugin),
            run_dir or self.project.run_dir(plugin.name),
        )

        self.plugins_service = plugins_service or ProjectPluginsService(project)
        self.settings_service = plugin_settings_service or PluginSettingsService(
            project, plugin, plugins_service=self.plugins_service
        )

        self._prepared = False
        self.plugin_config = {}
        self.plugin_config_processed = {}
        self.plugin_config_extras = {}
        self.plugin_config_env = {}

    @property
    def capabilities(self):
        # we want to make sure the capabilites are immutable from the `PluginInvoker` interface
        return frozenset(self.plugin.capabilities)

    @property
    def files(self):
        plugin_files = {**self.plugin.config_files, **self.plugin.output_files}

        return {
            _key: self.plugin_config_service.run_dir.joinpath(filename)
            for _key, filename in plugin_files.items()
        }

    def prepare(self, session):
        self.plugin_config = self.settings_service.as_dict(
            extras=False, session=session
        )
        self.plugin_config_processed = self.settings_service.as_dict(
            extras=False, process=True, session=session
        )
        self.plugin_config_extras = self.settings_service.as_dict(
            extras=True, session=session
        )
        self.plugin_config_env = self.settings_service.as_env(session=session)

        with self.plugin.trigger_hooks("configure", self, session):
            self.plugin_config_service.configure()
            self._prepared = True

    def cleanup(self):
        self.plugin_config = {}
        self.plugin_config_processed = {}
        self.plugin_config_extras = {}
        self.plugin_config_env = {}

        with self.plugin.trigger_hooks("cleanup", self):
            self._prepared = False

    @contextmanager
    def prepared(self, session):
        try:
            self.prepare(session)
            yield
        finally:
            self.cleanup()

    def exec_path(self):
        if not self.venv_service:
            if "/" not in self.plugin.executable.replace("\\", "/"):
                # Expect executable on path
                return self.plugin.executable

            # Return executable relative to project directory
            return self.project.root.joinpath(self.plugin.executable)

        # Return executable within venv
        return self.venv_service.exec_path(self.plugin.executable)

    def exec_args(self, *args, command=None, env=None):
        """Materialize the arguments to be passed to the executable."""
        env = env or {}
        if command:
            try:
                plugin_args = self.plugin.all_commands[command].expanded_args(
                    command, env
                )
            except KeyError as err:
                raise UnknownCommandError(self.plugin, command) from err
        else:
            plugin_args = self.plugin.exec_args(self)

        return [str(arg) for arg in (self.exec_path(), *plugin_args, *args)]

    def env(self):
        env = {
            **self.project.dotenv_env,
            **self.settings_service.env,
            **self.plugin_config_env,
        }

        # Ensure Meltano venv is not inherited
        env.pop("VIRTUAL_ENV", None)
        env.pop("PYTHONPATH", None)
        if self.venv_service:
            # Switch to plugin-specific venv
            venv = VirtualEnv(
                self.project.venvs_dir(self.plugin.type, self.plugin.name)
            )
            venv_dir = str(venv.bin_dir)
            env["VIRTUAL_ENV"] = str(venv.root)
            env["PATH"] = os.pathsep.join([venv_dir, env["PATH"]])

        return env

    def Popen_options(self):
        return {}

    @contextmanager
    def _invoke(
        self,
        *args,
        require_preparation=True,
        env=None,
        command=None,
        **kwargs,
    ):
        env = env or {}

        if require_preparation and not self._prepared:
            raise InvokerNotPreparedError()

        with self.plugin.trigger_hooks("invoke", self, args):
            popen_options = {**self.Popen_options(), **kwargs}
            popen_env = {**self.env(), **env}
            popen_args = self.exec_args(*args, command=command, env=popen_env)
            logging.debug(f"Invoking: {popen_args}")
            logging.debug(f"Env: {popen_env}")

            try:
                yield (popen_args, popen_options, popen_env)
            except FileNotFoundError as err:
                raise ExecutableNotFoundError(
                    self.plugin, self.plugin.executable
                ) from err

    def invoke(self, *args, **kwargs):
        with self._invoke(*args, **kwargs) as (popen_args, popen_options, popen_env):
            return subprocess.Popen(popen_args, **popen_options, env=popen_env)

    async def invoke_async(self, *args, **kwargs):
        with self._invoke(*args, **kwargs) as (popen_args, popen_options, popen_env):
            return await asyncio.create_subprocess_exec(
                *popen_args,
                **popen_options,
                env=popen_env,
            )

    def dump(self, file_id):
        try:
            with self._invoke():
                return self.files[file_id].read_text()
        except ExecutableNotFoundError as err:
            # Unwrap FileNotFoundError
            raise err.__cause__



import json
from pathlib import Path
from typing import Dict

from meltano.core.sql.design_helper import DesignHelper


class DesignMissingError(Exception):
    pass


class ConnectionNotFoundError(Exception):
    pass


class M5ocFile:
    def __init__(self, content: Dict):
        self.content = content

    @classmethod
    def load(cls, file_path):
        with file_path.open() as f:
            m5oc = M5ocFile(json.load(f))

        return m5oc

    @property
    def designs(self):
        return list(map(DesignHelper, self.content["designs"]))

    @property
    def reports(self):
        return self.content["reports"]

    def connection(self, connection_name: str) -> Dict:
        try:
            return self.content["connection"]
        except StopIteration:
            raise ConnectionNotFoundError(f"{connection_name} not found.")

    def design(self, design_name: str) -> Dict:
        try:
            return next(e for e in self.designs if e["name"] == design_name)
        except StopIteration:
            raise DesignMissingError(f"{design_name} not found.")

    def report(self, report_name: str) -> Dict:
        try:
            return next(e for e in self.reports if e["name"] == report_name)
        except StopIteration:
            raise DesignMissingError(f"{report_name} not found.")



import asyncio
import logging
import os
import re
import sys
from contextlib import suppress

from meltano.core.project_settings_service import ProjectSettingsService

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
DEFAULT_LEVEL = "info"
FORMAT = (
    "[%(asctime)s] [%(process)d|%(threadName)10s|%(name)s] [%(levelname)s] %(message)s"
)


def parse_log_level(log_level):
    return LEVELS.get(log_level, LEVELS[DEFAULT_LEVEL])


def setup_logging(project=None, log_level=DEFAULT_LEVEL):
    # Mimick Python 3.8's `force=True` kwarg to override any
    # existing logger handlers
    # See https://github.com/python/cpython/commit/cf67d6a934b51b1f97e72945b596477b271f70b8
    root = logging.getLogger()
    for h in root.handlers[:]:
        root.removeHandler(h)
        h.close()

    if project:
        settings_service = ProjectSettingsService(project)
        log_level = settings_service.get("cli.log_level")

    logging.basicConfig(
        stream=sys.stderr, format=FORMAT, level=parse_log_level(log_level)
    )


def remove_ansi_escape_sequences(line):
    """
    Remove ANSI escape sequences that are used for adding colors in
     terminals, so that only the text is logged in a file
    """
    ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")
    return ansi_escape.sub("", line)


async def _write_line_writer(writer, line):
    # StreamWriters like a subprocess's stdin need special consideration
    if isinstance(writer, asyncio.StreamWriter):
        try:  # noqa: WPS229
            writer.write(line)
            await writer.drain()
        except (BrokenPipeError, ConnectionResetError):
            with suppress(AttributeError):  # `wait_closed` is Python 3.7+
                await writer.wait_closed()

            return False
    else:
        writer.writeline(line.decode())

    return True


async def capture_subprocess_output(reader, *line_writers):
    """
    Capture in real time the output stream of a suprocess that is run async.
    The stream has been set to asyncio.subprocess.PIPE and is provided using
     reader to this function.
    As new lines are captured for reader, they are written to output_stream.
    This async function should be run with await asyncio.wait() while waiting
     for the subprocess to end.
    """
    while not reader.at_eof():
        line = await reader.readline()
        if not line:
            continue

        for writer in line_writers:
            if not await _write_line_writer(writer, line):
                # If the destination stream is closed, we can stop capturing output.
                return



import io
import logging
import os
import re
import shutil
from copy import deepcopy
from typing import Dict, Iterable, List, Optional

import meltano
import meltano.core.bundle as bundle
import requests
import yaml

from .behavior.canonical import Canonical
from .behavior.versioned import IncompatibleVersionError, Versioned
from .plugin import BasePlugin, PluginDefinition, PluginRef, PluginType, Variant
from .plugin.error import PluginNotFoundError
from .plugin.factory import base_plugin_factory
from .plugin.project_plugin import ProjectPlugin
from .project_settings_service import ProjectSettingsService
from .setting_definition import SettingDefinition
from .utils import NotFound, find_named


class DiscoveryInvalidError(Exception):
    """Occurs when the discovery.yml fails to be parsed."""

    pass


class DiscoveryUnavailableError(Exception):
    """Occurs when the discovery.yml cannot be found or downloaded."""

    pass


# Increment this version number whenever the schema of discovery.yml is changed.
# See https://www.meltano.com/docs/contributor-guide.html#discovery-yml-version for more information.
VERSION = 18


class DiscoveryFile(Canonical):
    def __init__(self, version=1, **plugins):
        super().__init__(version=int(version))

        for plugin_type in PluginType:
            self[plugin_type] = []

        for plugin_type, raw_plugins in plugins.items():
            for raw_plugin in raw_plugins:
                plugin_def = PluginDefinition(
                    plugin_type,
                    raw_plugin.pop("name"),
                    raw_plugin.pop("namespace"),
                    **raw_plugin,
                )
                self[plugin_type].append(plugin_def)

    @classmethod
    def file_version(cls, attrs):
        """Return version of discovery file represented by attrs dictionary."""
        return int(attrs.get("version", 1))


class PluginDiscoveryService(Versioned):
    __version__ = VERSION

    def __init__(self, project, discovery: Optional[Dict] = None):
        self.project = project

        self._discovery_version = None
        self._discovery = None
        if discovery:
            self._discovery_version = DiscoveryFile.file_version(discovery)
            self._discovery = DiscoveryFile.parse(discovery)

        self.settings_service = ProjectSettingsService(self.project)

    @property
    def file_version(self):
        return self._discovery_version

    @property
    def discovery_url(self):
        discovery_url = self.settings_service.get("discovery_url")

        if not discovery_url or not re.match(r"^https?://", discovery_url):
            return None

        return discovery_url

    @property
    def discovery(self):
        """
        Return first compatible discovery manifest from these locations:

        - project local `discovery.yml`
        - `discovery_url` project setting
        - .meltano/cache/discovery.yml
        - meltano.core.bundle
        """
        if self._discovery:
            return self._discovery

        loaders = [
            (
                self.load_local_discovery,
                "your project's local `discovery.yml` manifest",
            ),
            (
                self.load_remote_discovery,
                f"the `discovery.yml` manifest received from {self.discovery_url}",
            ),
            (self.load_cached_discovery, "the cached `discovery.yml` manifest"),
            (self.load_bundled_discovery, "the bundled `discovery.yml` manifest"),
        ]

        errored = False
        for loader, description in loaders:
            if errored:
                logging.warning(f"Falling back on {description}...")

            try:
                loader()
            except IncompatibleVersionError as err:
                errored = True

                logging.warning(
                    f"{description.capitalize()} has version {err.file_version}, while this version of Meltano requires version {err.version}."
                )
                if err.file_version > err.version:
                    logging.warning(
                        "Please install the latest compatible version of Meltano using `meltano upgrade`."
                    )
            except DiscoveryInvalidError as err:
                errored = True
                logging.error(f"{description.capitalize()} could not be parsed.")
                logging.debug(str(err))

            if self._discovery:
                return self._discovery

        raise DiscoveryInvalidError("No valid `discovery.yml` manifest could be found")

    def load_local_discovery(self):
        try:
            with self.project.root_dir("discovery.yml").open() as local_discovery:
                return self.load_discovery(local_discovery)
        except FileNotFoundError:
            pass

    def load_remote_discovery(self):
        discovery_url = self.discovery_url
        if not discovery_url:
            return

        try:
            headers = {"User-Agent": f"Meltano/{meltano.__version__}"}  # noqa: WPS609
            params = {}

            if self.settings_service.get("send_anonymous_usage_stats"):
                project_id = self.settings_service.get("project_id")

                headers["X-Project-ID"] = project_id
                params["project_id"] = project_id

            response = requests.get(discovery_url, headers=headers, params=params)
            response.raise_for_status()

            remote_discovery = io.StringIO(response.text)
            discovery = self.load_discovery(remote_discovery, cache=True)

            return discovery
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
        ) as err:
            logging.debug("Remote `discovery.yml` manifest could not be downloaded.")
            logging.debug(str(err))
            pass

    def load_cached_discovery(self):
        try:
            with self.cached_discovery_file.open() as cached_discovery:
                return self.load_discovery(cached_discovery)
        except FileNotFoundError:
            pass

    def load_bundled_discovery(self):
        with bundle.find("discovery.yml").open() as bundled_discovery:
            discovery = self.load_discovery(bundled_discovery, cache=True)

        return discovery

    def load_discovery(self, discovery_file, cache=False):
        try:
            discovery_yaml = yaml.safe_load(discovery_file)

            self._discovery_version = DiscoveryFile.file_version(discovery_yaml)
            self.ensure_compatible()

            self._discovery = DiscoveryFile.parse(discovery_yaml)

            if cache:
                self.cache_discovery()

            return self._discovery
        except IncompatibleVersionError:
            raise
        except (yaml.YAMLError, Exception) as err:
            raise DiscoveryInvalidError(str(err))

    def cache_discovery(self):
        with self.cached_discovery_file.open("w") as cached_discovery:
            yaml.dump(
                self._discovery,
                cached_discovery,
                default_flow_style=False,
                sort_keys=False,
            )

    @property
    def cached_discovery_file(self):
        return self.project.meltano_dir("cache", "discovery.yml")

    def get_plugins_of_type(self, plugin_type):
        return self.discovery[plugin_type]

    def plugins_by_type(self):
        return {
            plugin_type: self.get_plugins_of_type(plugin_type)
            for plugin_type in PluginType
        }

    def plugins(self) -> Iterable[PluginDefinition]:
        yield from (
            plugin
            for plugin_type, plugins in self.plugins_by_type().items()
            for plugin in plugins
        )

    def find_definition(
        self, plugin_type: PluginType, plugin_name: str
    ) -> PluginDefinition:
        try:
            return find_named(self.get_plugins_of_type(plugin_type), plugin_name)
        except NotFound as err:
            raise PluginNotFoundError(PluginRef(plugin_type, plugin_name)) from err

    def find_definition_by_namespace(
        self, plugin_type: PluginType, namespace: str
    ) -> PluginDefinition:
        try:
            return next(
                plugin
                for plugin in self.get_plugins_of_type(plugin_type)
                if plugin.namespace == namespace
            )
        except StopIteration as stop:
            raise PluginNotFoundError(namespace) from stop

    def find_base_plugin(
        self, plugin_type: PluginType, plugin_name: str, variant=None
    ) -> BasePlugin:
        plugin = self.find_definition(plugin_type, plugin_name)
        return base_plugin_factory(plugin, variant)

    def get_base_plugin(self, project_plugin: ProjectPlugin) -> BasePlugin:
        plugin = project_plugin.custom_definition or self.find_definition(
            project_plugin.type, project_plugin.inherit_from or project_plugin.name
        )

        return base_plugin_factory(plugin, project_plugin.variant)

    def find_related_plugin_refs(
        self,
        target_plugin: ProjectPlugin,
        plugin_types: List[PluginType] = list(PluginType),
    ):
        try:
            plugin_types.remove(target_plugin.type)
        except ValueError:
            pass

        related_plugin_refs = []

        runner_ref = target_plugin.runner
        if runner_ref:
            related_plugin_refs.append(runner_ref)

        related_plugin_refs.extend(
            related_plugin_def
            for plugin_type in plugin_types
            for related_plugin_def in self.get_plugins_of_type(plugin_type)
            if related_plugin_def.namespace == target_plugin.namespace
        )

        return related_plugin_refs



import atexit
import datetime
import importlib
import logging
import logging.handlers
import os
from urllib.parse import urlsplit

import meltano.api.config
from flask import Flask, g, jsonify, request
from flask_cors import CORS
from flask_login import current_user
from meltano.api.headers import *
from meltano.api.security.auth import HTTP_READONLY_CODE
from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.core.db import project_engine
from meltano.core.logging.utils import FORMAT, setup_logging
from meltano.core.project import Project, ProjectReadonly
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.oauth.app import create_app as create_oauth_service
from werkzeug.wsgi import DispatcherMiddleware

setup_logging()

# the file logger we will set up below is for the `meltano.api` module
logger = logging.getLogger("meltano.api")


def create_app(config={}):
    project = Project.find()
    setup_logging(project)

    settings_service = ProjectSettingsService(project)

    project_engine(project, default=True)

    app = Flask(
        __name__, instance_path=str(project.root), instance_relative_config=True
    )

    # make sure we have the latest environment loaded
    importlib.reload(meltano.api.config)

    app.config.from_object("meltano.api.config")
    app.config.from_mapping(**meltano.api.config.ProjectSettings(project).as_dict())
    app.config.from_mapping(**config)

    # File logging
    file_handler = logging.handlers.RotatingFileHandler(
        str(project.run_dir("meltano-ui.log")), backupCount=3
    )
    formatter = logging.Formatter(fmt=FORMAT)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    # 1) Extensions
    security_options = {}

    from .models import db
    from .mail import mail
    from .executor import setup_executor
    from .security import security, users, setup_security
    from .security.oauth import setup_oauth
    from .json import setup_json

    db.init_app(app)
    mail.init_app(app)
    setup_executor(app, project)
    setup_security(app, project)
    setup_oauth(app)
    setup_json(app)

    # we need to setup CORS for development
    if app.env == "development":
        CORS(app, origins="http://localhost:8080", supports_credentials=True)

    # 2) Register the URL Converters
    from .url_converters import PluginRefConverter

    app.url_map.converters["plugin_ref"] = PluginRefConverter

    # 3) Register the controllers

    from .controllers.dashboards import dashboardsBP
    from .controllers.embeds import embedsBP
    from .controllers.reports import reportsBP
    from .controllers.repos import reposBP
    from .controllers.settings import settingsBP
    from .controllers.sql import sqlBP
    from .controllers.orchestrations import orchestrationsBP
    from .controllers.plugins import pluginsBP
    from .controllers.root import root, api_root

    app.register_blueprint(dashboardsBP)
    app.register_blueprint(embedsBP)
    app.register_blueprint(reportsBP)
    app.register_blueprint(reposBP)
    app.register_blueprint(settingsBP)
    app.register_blueprint(sqlBP)
    app.register_blueprint(orchestrationsBP)
    app.register_blueprint(pluginsBP)
    app.register_blueprint(root)
    app.register_blueprint(api_root)

    if app.config["PROFILE"]:
        from .profiler import init

        init(app)

    # Notifications
    if settings_service.get("ui.notification"):
        from .events import notifications

        notifications.init_app(app)
        logger.debug("Notifications are enabled.")
    else:
        logger.debug("Notifications are disabled.")

    # Google Analytics setup
    tracker = GoogleAnalyticsTracker(project)

    @app.before_request
    def setup_js_context():
        # setup the appUrl
        appUrl = urlsplit(request.host_url)
        g.jsContext = {"appUrl": appUrl.geturl()[:-1], "version": meltano.__version__}

        setting_map = {
            "isSendAnonymousUsageStats": "send_anonymous_usage_stats",
            "projectId": "project_id",
            "trackingID": "tracking_ids.ui",
            "embedTrackingID": "tracking_ids.ui_embed",
            "isProjectReadonlyEnabled": "project_readonly",
            "isReadonlyEnabled": "ui.readonly",
            "isAnonymousReadonlyEnabled": "ui.anonymous_readonly",
            "isNotificationEnabled": "ui.notification",
            "isAnalysisEnabled": "ui.analysis",
            "logoUrl": "ui.logo_url",
            "oauthServiceUrl": "oauth_service.url",
        }

        for context_key, setting_name in setting_map.items():
            g.jsContext[context_key] = settings_service.get(setting_name)

        providers = settings_service.get("oauth_service.providers")
        g.jsContext["oauthServiceProviders"] = [
            provider for provider in providers.split(",") if provider
        ]

    @app.after_request
    def after_request(res):
        res.headers[VERSION_HEADER] = meltano.__version__
        return res

    @app.errorhandler(500)
    def internal_error(exception):
        logger.info(f"Error: {exception}")
        return jsonify({"error": True, "code": str(exception)}), 500

    @app.errorhandler(ProjectReadonly)
    def _handle(ex):
        return (jsonify({"error": True, "code": str(ex)}), HTTP_READONLY_CODE)

    # create the dispatcher to host the `OAuthService`
    app.wsgi_app = DispatcherMiddleware(
        app.wsgi_app, {"/-/oauth": create_oauth_service()}
    )

    return app



import json
import logging
import os
from typing import List

import yaml

from .plugin import BasePlugin, PluginType, Variant
from .plugin.project_plugin import ProjectPlugin
from .project import Project
from .project_plugins_service import PluginAlreadyAddedException, ProjectPluginsService


class MissingPluginException(Exception):
    pass


class ProjectAddService:
    def __init__(self, project: Project, plugins_service: ProjectPluginsService = None):
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)

    def add(self, plugin_type: PluginType, plugin_name: str, **attrs) -> ProjectPlugin:
        """Add plugin to project."""
        plugin = ProjectPlugin(
            plugin_type, plugin_name, **attrs, default_variant=Variant.DEFAULT_NAME
        )

        self.plugins_service.ensure_parent(plugin)

        # If we are inheriting from a base plugin definition,
        # repeat the variant and pip_url in meltano.yml
        parent = plugin.parent
        if isinstance(parent, BasePlugin):
            plugin.variant = parent.variant
            plugin.pip_url = parent.pip_url

        return self.add_plugin(plugin)

    def add_plugin(self, plugin: ProjectPlugin):
        return self.plugins_service.add_to_file(plugin)

    def add_related(self, *args, **kwargs):
        related_plugin_refs = (
            self.plugins_service.discovery_service.find_related_plugin_refs(
                *args, **kwargs
            )
        )

        added_plugins = []
        for plugin_ref in related_plugin_refs:
            try:
                plugin = self.add(plugin_ref.type, plugin_ref.name)
            except PluginAlreadyAddedException as err:
                continue

            added_plugins.append(plugin)

        added_plugins_with_related = []
        for plugin in added_plugins:
            added_plugins_with_related.extend(
                [plugin, *self.add_related(plugin, **kwargs)]
            )

        return added_plugins_with_related



import importlib
import logging
import os
import shutil
import signal
import subprocess
from typing import Optional

import click
import meltano
import meltano.core.bundle as bundle
import psutil
from meltano.cli.utils import PluginInstallReason, install_plugins
from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.core.migration_service import MigrationError, MigrationService
from meltano.core.project import Project
from meltano.core.project_plugins_service import PluginType, ProjectPluginsService


class UpgradeError(Exception):
    """Occurs when the Meltano upgrade fails"""

    pass


class AutomaticPackageUpgradeError(Exception):
    def __init__(self, reason, instructions):
        self.reason = reason
        self.instructions = instructions


class UpgradeService:
    def __init__(self, engine, project: Project):
        self.project = project
        self.engine = engine

    def reload_ui(self):
        click.secho("Reloading UI...", fg="blue")

        pid_file_path = self.project.run_dir("gunicorn.pid")
        try:
            with pid_file_path.open("r") as pid_file:
                pid = int(pid_file.read())

                process = psutil.Process(pid)
                process.send_signal(signal.SIGHUP)
        except FileNotFoundError:
            click.echo("UI is not running")
        except Exception as ex:
            logging.error(f"Cannot restart from `{pid_file_path}`: {ex}")

    def _upgrade_package(self, pip_url: Optional[str] = None, force=False):
        meltano_file_path = "/src/meltano/__init__.py"
        editable = meltano.__file__.endswith(meltano_file_path)
        if editable and not force:
            meltano_dir = meltano.__file__[0 : -len(meltano_file_path)]
            raise AutomaticPackageUpgradeError(
                reason="it is installed from source",
                instructions=f"navigate to `{meltano_dir}` and run `git pull`",
            )

        in_docker = os.path.exists("/.dockerenv")
        if in_docker:
            raise AutomaticPackageUpgradeError(
                reason="it is installed inside Docker",
                instructions="pull the latest Docker image using `docker pull meltano/meltano` and recreate any containers you may have created",
            )

        pip_url = pip_url or "meltano"
        run = subprocess.run(
            ["pip", "install", "--upgrade", pip_url],
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if run.returncode != 0:
            raise UpgradeError(f"Failed to upgrade `meltano`.", run)

        return True

    def upgrade_package(self, *args, **kwargs):
        click.secho("Upgrading `meltano` package...", fg="blue")

        try:
            self._upgrade_package(*args, **kwargs)
            click.echo("The `meltano` package has been upgraded.")

            click.echo()
            self.reload_ui()

            return True
        except AutomaticPackageUpgradeError as err:
            click.echo(
                f"{click.style('The `meltano` package could not be upgraded automatically', fg='red')} because {err.reason}."
            )
            click.echo(f"To upgrade manually, {err.instructions}.")
            return False

    def update_files(self):
        """
        Update the files managed by Meltano inside the current project.
        """
        click.secho("Updating files managed by plugins...", fg="blue")

        file_plugins = ProjectPluginsService(self.project).get_plugins_of_type(
            PluginType.FILES
        )
        if not file_plugins:
            click.echo("Nothing to update")
            return

        install_plugins(self.project, file_plugins, reason=PluginInstallReason.UPGRADE)

    def migrate_database(self):
        click.secho("Applying migrations to system database...", fg="blue")

        try:
            migration_service = MigrationService(self.engine)
            migration_service.upgrade()
            migration_service.seed(self.project)
        except MigrationError as err:
            raise UpgradeError(str(err)) from err

    def compile_models(self):
        click.secho("Recompiling models...", fg="blue")

        ProjectCompiler(self.project).compile()

    def upgrade(self, skip_package=False, **kwargs):
        package_upgraded = False
        if not skip_package:
            package_upgraded = self.upgrade_package(**kwargs)

            if not package_upgraded:
                click.echo(
                    "Then, run `meltano upgrade --skip-package` to upgrade your project based on the latest version."
                )
                return

            click.echo()

        self.update_files()
        click.echo()
        self.migrate_database()
        click.echo()
        self.compile_models()

        click.echo()
        if package_upgraded:
            click.secho(
                "Meltano and your Meltano project have been upgraded!", fg="green"
            )
        else:
            click.secho("Your Meltano project has been upgraded!", fg="green")



import copy
from typing import Iterable

import yaml


class Canonical:
    """
    This class defines an object that can be reprensented as a subset of
    its attributes.

    Its purpose is to be serializable as the smallest possible form.

    The attribute rules are:
      - All attributes that are Truthy
      - All attributes that are boolean (False is valid)
      - All attributes that are listed in the `_verbatim` set and non-null
      - All attributes that start with "_" are excluded
    """

    def __init__(self, *args, **attrs):
        self._dict = {}

        super().__init__(*args)

        for attr, value in attrs.items():
            setattr(self, attr, value)

        self._verbatim = set()
        self._flattened = {"extras"}

        self._fallback_to = None
        self._fallbacks = set()
        self._defaults = {}

    @classmethod
    def as_canonical(cls, target):
        if isinstance(target, Canonical):
            return {key: Canonical.as_canonical(val) for key, val in target}

        if isinstance(target, set):
            return list(map(Canonical.as_canonical, target))

        if isinstance(target, list):
            return list(map(Canonical.as_canonical, target))

        if isinstance(target, dict):
            results = {}
            for key, val in target.items():
                if isinstance(val, Canonical):
                    results[key] = val.canonical()
                else:
                    results[key] = Canonical.as_canonical(val)
            return results

        return copy.deepcopy(target)

    def canonical(self):
        return Canonical.as_canonical(self)

    def with_attrs(self, *args, **kwargs):
        return self.__class__(**{**self.canonical(), **kwargs})

    @classmethod
    def parse(cls, obj) -> "Canonical":
        if obj is None:
            return None

        if isinstance(obj, Canonical):
            return obj

        return cls(**obj)

    def is_attr_set(self, attr):
        """Return whether specified attribute has a non-default/fallback value set."""
        return self._dict.get(attr) is not None

    def __getattr__(self, attr):
        try:
            value = self._dict[attr]
        except KeyError as err:
            if self._fallback_to and not attr.startswith("_"):
                return getattr(self._fallback_to, attr)

            raise AttributeError(attr) from err

        if value is not None:
            return value

        if attr in self._fallbacks and self._fallback_to:
            value = getattr(self._fallback_to, attr)

        if value is not None:
            return value

        if attr in self._defaults:
            value = self._defaults[attr](self)

        return value

    def __setattr__(self, attr, value):
        if attr.startswith("_") or hasattr(self.__class__, attr):
            super().__setattr__(attr, value)
        else:
            self._dict[attr] = value

    def __getitem__(self, attr):
        return getattr(self, attr)

    def __setitem__(self, attr, value):
        return setattr(self, attr, value)

    def __iter__(self):
        for k, v in self._dict.items():
            if not v:
                if k in self._verbatim:
                    if v is None:
                        continue
                else:
                    # bool values are valid and should be forwarded
                    if v is not False:
                        continue

            # empty canonicals should be skipped
            if isinstance(v, Canonical) and not dict(v):
                continue

            if k in self._flattened:
                if isinstance(v, Canonical):
                    yield from v
                else:
                    yield from v.items()
            else:
                yield (k, v)

    def __len__(self):
        return len(self._dict)

    def __contains__(self, obj):
        return obj in self._dict

    def update(self, *others, **kwargs):
        if kwargs:
            others = [*others, kwargs]

        for other in others:
            other = Canonical.as_canonical(other)

            for k, v in other.items():
                setattr(self, k, v)

    @classmethod
    def yaml(cls, dumper, obj):
        return dumper.represent_mapping(
            "tag:yaml.org,2002:map", Canonical.as_canonical(obj), flow_style=False
        )


yaml.add_multi_representer(Canonical, Canonical.yaml)



import collections
from enum import Enum

from .base import MeltanoFilterExpressionType

FilterOption = collections.namedtuple("FilterOption", "label description expression")


FilterOptions = [
    FilterOption(
        label="Less than",
        description="Less than",
        expression=MeltanoFilterExpressionType.LessThan,
    ),
    FilterOption(
        label="Less than or equal",
        description="Less than or equal",
        expression=MeltanoFilterExpressionType.LessOrEqualThan,
    ),
    FilterOption(
        label="Equal to",
        description="Equal to",
        expression=MeltanoFilterExpressionType.EqualTo,
    ),
    FilterOption(
        label="Greater than or equal",
        description="Greater than or equal",
        expression=MeltanoFilterExpressionType.GreaterOrEqualThan,
    ),
    FilterOption(
        label="Greater than",
        description="Greater than",
        expression=MeltanoFilterExpressionType.GreaterThan,
    ),
    FilterOption(
        label="Like",
        description="Custom like expression",
        expression=MeltanoFilterExpressionType.Like,
    ),
    FilterOption(
        label="Is Null",
        description="Is null",
        expression=MeltanoFilterExpressionType.IsNull,
    ),
    FilterOption(
        label="Is Not Null",
        description="Is not null",
        expression=MeltanoFilterExpressionType.IsNotNull,
    ),
]



"""Add `last_heartbeat_at` column to `job` table.

Revision ID: d135f52a6f49
Revises: e4fbabc3fed6
Create Date: 2021-01-14 12:45:55.821947

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d135f52a6f49"
down_revision = "e4fbabc3fed6"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("job", sa.Column("last_heartbeat_at", sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column("user", "last_heartbeat_at")



"""Defines MeltanoInvoker."""
import os
import subprocess
import sys
from pathlib import Path

from .project import Project
from .project_settings_service import ProjectSettingsService, SettingValueStore

MELTANO_COMMAND = "meltano"


class MeltanoInvoker:
    def __init__(self, project, settings_service: ProjectSettingsService = None):
        self.project = project
        self.settings_service = settings_service or ProjectSettingsService(project)

    def invoke(self, args, command=MELTANO_COMMAND, env=None, **kwargs):
        """Invoke `meltano` (or provided command) with provided args and env."""
        return subprocess.run(
            [self._executable_path(command), *args],
            **kwargs,
            env=self._executable_env(env)
        )

    def _executable_path(self, command):
        if command == MELTANO_COMMAND:
            # This symlink is created by Project.activate
            executable_symlink = self.project.run_dir().joinpath("bin")
            if executable_symlink.exists():
                return str(executable_symlink)

        executable = Path(os.path.dirname(sys.executable), command)
        if executable.exists():
            return str(executable)

        # Fall back on expecting command to be in the PATH
        return command

    def _executable_env(self, env=None):
        exec_env = {}

        # Include env that project settings are evaluated in
        exec_env.update(self.settings_service.env)

        # Include env for settings explicitly overridden using CLI flags
        exec_env.update(
            self.settings_service.as_env(source=SettingValueStore.CONFIG_OVERRIDE)
        )

        # Include explicitly provided env
        if env:
            exec_env.update(env)

        return exec_env



import asyncio
import logging
import os
import secrets
import signal
import subprocess

import click
from click_default_group import DefaultGroup
from meltano.api.workers import APIWorker, MeltanoCompilerWorker, UIAvailableWorker
from meltano.core.db import project_engine
from meltano.core.migration_service import MigrationService
from meltano.core.project_settings_service import (
    ProjectSettingsService,
    SettingValueStore,
)
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.utils import truthy

from . import cli
from .params import pass_project
from .utils import CliError

logger = logging.getLogger(__name__)


def ensure_secure_setup(project):
    settings_service = ProjectSettingsService(project)

    if not settings_service.get("ui.authentication"):
        return

    facts = []
    if (
        settings_service.get("ui.server_name") is None
        and settings_service.get("ui.session_cookie_domain") is None
    ):
        facts.append(
            f"- Neither the 'ui.server_name' or 'ui.session_cookie_domain' setting has been set"
        )

    secure_settings = ["ui.secret_key", "ui.password_salt"]
    for setting_name in secure_settings:
        value, source = settings_service.get_with_source(setting_name)
        if source is SettingValueStore.DEFAULT:
            facts.append(
                f"- The '{setting_name}' setting has not been changed from the default test value"
            )

    if facts:
        click.secho(
            "Authentication is enabled, but your configuration is currently insecure:",
            fg="red",
        )
        for fact in facts:
            click.echo(fact)
        click.echo(
            "For more information about these settings and how to set them, visit https://www.meltano.com/docs/settings.html#ui-authentication"
        )
        click.echo()


def start_workers(workers):
    def stop_all():
        logger.info("Stopping all background workers...")
        for worker in workers:
            worker.stop()

    # start all workers
    for worker in workers:
        worker.start()

    return stop_all


@cli.group(cls=DefaultGroup, default="start", default_if_no_args=True)
@pass_project(migrate=True)
@click.pass_context
def ui(ctx, project):
    ctx.obj["project"] = project


@ui.command()
@click.option("--reload", is_flag=True, default=False)
@click.option("--bind", help="The hostname (or IP address) to bind on")
@click.option("--bind-port", help="Port to run webserver on", type=int)
@click.pass_context
def start(ctx, reload, bind, bind_port):
    if bind:
        ProjectSettingsService.config_override["ui.bind_host"] = bind
    if bind_port:
        ProjectSettingsService.config_override["ui.bind_port"] = bind_port

    project = ctx.obj["project"]
    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_ui()

    ensure_secure_setup(project)

    workers = []

    try:
        compiler_worker = MeltanoCompilerWorker(project)
        compiler_worker.compiler.compile()
        workers.append(compiler_worker)
    except Exception as e:
        logger.error(f"Initial compilation failed: {e}")

    workers.append(UIAvailableWorker(project))
    workers.append(
        APIWorker(project, reload=reload or os.getenv("FLASK_ENV") == "development")
    )

    cleanup = start_workers(workers)

    def handle_terminate(signal, frame):
        cleanup()

    signal.signal(signal.SIGTERM, handle_terminate)
    logger.info("All workers started.")


@ui.command()
@click.argument("server_name")
@click.option("--bits", default=256)
@click.pass_context
def setup(ctx, server_name, **flags):
    """
    Generates and stores server name and secrets.
    """
    project = ctx.obj["project"]
    settings_service = ProjectSettingsService(project)

    def set_setting_env(setting_name, value):
        settings_service.set(setting_name, value, store=SettingValueStore.DOTENV)

    set_setting_env("ui.server_name", server_name)

    ui_cfg_path = project.root_dir("ui.cfg")
    if ui_cfg_path.exists():
        raise CliError(
            f"Found existing secrets in file '{ui_cfg_path}'. Please delete this file and rerun this command to regenerate the secrets."
        )

    generate_secret = lambda: secrets.token_hex(int(flags["bits"] / 8))  # in bytes

    secret_settings = ["ui.secret_key", "ui.password_salt"]
    for setting_name in secret_settings:
        value, source = settings_service.get_with_source(setting_name)
        if source is not SettingValueStore.DEFAULT:
            click.echo(
                f"Setting '{setting_name}' has already been set in {source.label}. Please unset it manually and rerun this command to regenerate this secret."
            )
        else:
            set_setting_env(setting_name, generate_secret())

    click.echo(
        "The server name and generated secrets have been stored in your project's `.env` file."
    )
    click.echo(
        "In production, you will likely want to move these settings to actual environment variables, since `.env` is in `.gitignore` by default."
    )



"""Defines PluginRemoveService."""
from typing import Iterable, Tuple

from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_location_remove import (
    DbRemoveManager,
    InstallationRemoveManager,
    MeltanoYmlRemoveManager,
    PluginLocationRemoveStatus,
)
from meltano.core.project_plugins_service import ProjectPluginsService

from .project import Project
from .utils import noop


class PluginRemoveService:
    """Handle plugin installation removal operations."""

    def __init__(self, project: Project, plugins_service: ProjectPluginsService = None):
        """Construct a PluginRemoveService instance."""
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)

    def remove_plugins(
        self,
        plugins: Iterable[ProjectPlugin],
        plugin_status_cb=noop,
        removal_manager_status_cb=noop,
    ) -> Tuple[int, int]:
        """
        Remove multiple plugins.

        Returns a tuple containing:
        1. The total number of removed plugins
        2. The total number of plugins attempted
        """
        num_plugins: int = len(plugins)
        removed_plugins: int = num_plugins

        for plugin in plugins:
            plugin_status_cb(plugin)

            removal_managers = self.remove_plugin(plugin)

            any_not_removed = False
            for manager in removal_managers:
                any_not_removed = not manager.plugin_removed or any_not_removed
                removal_manager_status_cb(manager)

            if any_not_removed:
                removed_plugins -= 1

        return removed_plugins, num_plugins

    def remove_plugin(self, plugin: ProjectPlugin) -> Tuple[PluginLocationRemoveStatus]:
        """
        Remove a plugin from `meltano.yml`, its installation in `.meltano`, and its settings in the Meltano system database.

        Returns a tuple containing a remove manager for each location.
        """
        remove_managers = (
            DbRemoveManager(plugin, self.project),
            MeltanoYmlRemoveManager(plugin, self.project),
            InstallationRemoveManager(plugin, self.project),
        )

        for manager in remove_managers:
            manager.remove()

        return remove_managers



"""Defines Job model class."""
import asyncio
import logging
import os
import signal
import uuid
from contextlib import contextmanager, suppress
from datetime import datetime, timedelta
from enum import Enum

import sqlalchemy.types as types
from async_generator import asynccontextmanager
from meltano.core.error import Error
from meltano.core.models import SystemModel
from meltano.core.sqlalchemy import GUID, IntFlag, JSONEncodedDict
from sqlalchemy import Column
from sqlalchemy.ext.mutable import MutableDict

HEARTBEATLESS_JOB_VALID_HOURS = 24
HEARTBEAT_VALID_MINUTES = 5


class InconsistentStateError(Error):
    """
    Occur upon a wrong operation for the current state.
    """


class ImpossibleTransitionError(Error):
    """
    Occur upon a wrong transition.
    """


class State(Enum):
    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())

    def transitions(self):
        return self.value[1]

    def __str__(self):
        return self.name


def current_trigger():
    return os.getenv("MELTANO_JOB_TRIGGER")


class Payload(IntFlag):
    STATE = 1
    INCOMPLETE_STATE = 2


class Job(SystemModel):  # noqa: WPS214
    """Model class that represents a `meltano elt` run in the system database."""

    __tablename__ = "job"

    id = Column(types.Integer, primary_key=True)
    job_id = Column(types.String)
    run_id = Column(GUID, nullable=False, default=uuid.uuid4)
    state = Column(types.Enum(State, name="job_state"))
    started_at = Column(types.DateTime)
    last_heartbeat_at = Column(types.DateTime)
    ended_at = Column(types.DateTime)
    payload = Column(MutableDict.as_mutable(JSONEncodedDict))
    payload_flags = Column(IntFlag, default=0)
    trigger = Column(types.String, default=current_trigger)

    def __init__(self, **kwargs):
        kwargs["state"] = kwargs.get("state", State.IDLE)
        kwargs["payload"] = kwargs.get("payload", {})
        kwargs["run_id"] = kwargs.get("run_id", uuid.uuid4())
        super().__init__(**kwargs)

    def is_running(self):
        return self.state is State.RUNNING

    def is_stale(self):
        """
        Return whether job has gone stale.

        Running jobs with a heartbeat are considered stale after no heartbeat is recorded for 5 minutes.
        Legacy jobs without a heartbeat are considered stale after being in the running state for 24 hours.
        """
        if not self.is_running():
            return False

        if self.last_heartbeat_at:
            timestamp = self.last_heartbeat_at
            valid_for = timedelta(minutes=HEARTBEAT_VALID_MINUTES)
        else:
            timestamp = self.started_at
            valid_for = timedelta(hours=HEARTBEATLESS_JOB_VALID_HOURS)

        return datetime.utcnow() - timestamp > valid_for

    def has_error(self):
        return self.state is State.FAIL

    def is_complete(self):
        return self.state in [State.SUCCESS, State.FAIL]

    def is_success(self):
        return self.state is State.SUCCESS

    def can_transit(self, state: State) -> bool:
        if self.state is state:
            return True

        return state.name in self.state.transitions()

    def transit(self, state: State) -> (State, State):
        transition = (self.state, state)

        if not self.can_transit(state):
            raise ImpossibleTransitionError(transition)

        if self.state is state:
            return transition

        self.state = state

        return transition

    @asynccontextmanager
    async def run(self, session):
        """
        Run wrapped code in context of a job.

        Transitions state to RUNNING and SUCCESS/FAIL as appropriate and records heartbeat every second.
        """
        try:
            self.start()
            self.save(session)

            with self._handling_sigterm(session):
                async with self._heartbeating(session):
                    yield

            self.success()
            self.save(session)
        except BaseException as err:
            if not self.is_running():
                raise

            self.fail(error=self._error_message(err))
            self.save(session)

            raise

    def start(self):
        self.started_at = datetime.utcnow()
        self.transit(State.RUNNING)

    def fail(self, error=None):
        self.ended_at = datetime.utcnow()
        self.transit(State.FAIL)
        if error:
            self.payload.update({"error": str(error)})

    def success(self):
        self.ended_at = datetime.utcnow()
        self.transit(State.SUCCESS)

    def fail_stale(self):
        """Mark job as failed if it's gone stale."""
        if not self.is_stale():
            return False

        if self.last_heartbeat_at:
            reason = f"No heartbeat recorded for {HEARTBEAT_VALID_MINUTES} minutes."
        else:
            reason = f"Still running after {HEARTBEATLESS_JOB_VALID_HOURS} hours."

        self.fail(f"{reason} The process was likely killed unceremoniously.")

        return True

    def __repr__(self):
        return (
            "<Job(id='%s', job_id='%s', state='%s', started_at='%s', ended_at='%s')>"
            % (self.id, self.job_id, self.state, self.started_at, self.ended_at)
        )

    def save(self, session):
        session.add(self)
        session.commit()

        return self

    def _heartbeat(self):
        self.last_heartbeat_at = datetime.utcnow()

    async def _heartbeater(self, session):
        while True:
            self._heartbeat()
            self.save(session)

            await asyncio.sleep(1)

    @asynccontextmanager
    async def _heartbeating(self, session):
        heartbeat_future = asyncio.ensure_future(self._heartbeater(session))
        try:
            yield
        finally:
            if not heartbeat_future.cancelled():
                heartbeat_future.cancel()

            with suppress(asyncio.CancelledError):
                await heartbeat_future

    @contextmanager
    def _handling_sigterm(self, session):
        def handler(*_):  # noqa: WPS430
            sigterm_status = 143
            raise SystemExit(sigterm_status)

        original_termination_handler = signal.signal(signal.SIGTERM, handler)

        try:
            yield
        finally:
            signal.signal(signal.SIGTERM, original_termination_handler)

    def _error_message(self, err):
        if isinstance(err, SystemExit):
            return "The process was terminated"

        if isinstance(err, (KeyboardInterrupt, asyncio.CancelledError)):
            return "The process was interrupted"

        if str(err):
            return str(err)

        return repr(err)



import copy
import logging
from typing import Optional

from meltano.core.setting_definition import SettingDefinition
from meltano.core.utils import flatten, uniques_in

from .base import PluginDefinition, PluginRef, PluginType, Variant
from .command import Command
from .factory import base_plugin_factory

logger = logging.getLogger(__name__)


class CyclicInheritanceError(Exception):
    """Exception raised when project plugin inherits from itself cyclicly."""

    def __init__(self, plugin: "ProjectPlugin", ancestor: "ProjectPlugin"):
        """Initialize cyclic inheritance error."""
        self.plugin = plugin
        self.ancestor = ancestor

    def __str__(self):
        """Return error message."""
        return "{type} '{name}' cannot inherit from '{ancestor}', which itself inherits from '{name}'".format(
            type=self.plugin.type.descriptor.capitalize(),
            name=self.plugin.name,
            ancestor=self.ancestor.name,
        )


class ProjectPlugin(PluginRef):
    VARIANT_ATTR = "variant"

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        inherit_from: Optional[str] = None,
        namespace: Optional[str] = None,
        variant: Optional[str] = None,
        pip_url: Optional[str] = None,
        config: Optional[dict] = None,
        commands: Optional[dict] = None,
        default_variant=Variant.ORIGINAL_NAME,
        **extras,
    ):
        super().__init__(plugin_type, name)

        # Attributes will be listed in meltano.yml in the order they are set on self:
        self.inherit_from = (
            inherit_from if inherit_from and inherit_from != name else None
        )

        # If a custom definition is provided, its properties will come before all others in meltano.yml
        self.custom_definition = None
        self._flattened.add("custom_definition")

        self._parent = None
        if not self.inherit_from and namespace:
            # When not explicitly inheriting, a namespace indicates an embedded custom plugin definition
            self.custom_definition = PluginDefinition(
                plugin_type, name, namespace, variant=variant, pip_url=pip_url, **extras
            )

            # Any properties considered "extra" by the embedded plugin definition
            # should be considered extras of the project plugin, since they are
            # the current values, not default values.
            extras = self.custom_definition.extras
            self.custom_definition.extras = {}

            # Typically, the parent is set from ProjectPluginsService.current_plugins,
            # where we have access to the discoverable plugin definitions coming from
            # PluginDiscoveryService, but here we can set the parent directly.
            self.parent = base_plugin_factory(self.custom_definition, variant)

        # These properties are also set on the parent, but can be overridden
        self.namespace = namespace
        self.set_presentation_attrs(extras)
        self.variant = variant
        self.pip_url = pip_url
        self.commands = Command.parse_all(commands)

        self._fallbacks.update(
            ["logo_url", "description", self.VARIANT_ATTR, "pip_url"]
        )

        # If no variant is set, we fall back on the default
        self._defaults[self.VARIANT_ATTR] = lambda _: default_variant

        if self.inherit_from:
            # When explicitly inheriting from a project plugin or discoverable definition,
            # derive default values from our own name
            self._defaults["namespace"] = lambda plugin: plugin.name.replace("-", "_")
            self._defaults["label"] = lambda plugin: (
                f"{plugin.parent.label}: {plugin.name}"
                if plugin.parent
                else plugin.name
            )
        else:
            # When shadowing a discoverable definition with the same name (no `inherit_from`),
            # or an embedded custom definition (with `namespace`), fall back on parent's
            # values derived from its name instead
            self._fallbacks.update(["namespace", "label"])

        self.config = copy.deepcopy(config or {})
        self.extras = extras

        if "profiles" in extras:
            logger.warning(
                f"Plugin configuration profiles are no longer supported, ignoring `profiles` in '{name}' {plugin_type.descriptor} definition."
            )

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, new_parent):
        ancestor = new_parent
        while isinstance(ancestor, self.__class__):
            if ancestor == self:
                raise CyclicInheritanceError(self, ancestor)

            ancestor = ancestor.parent

        self._parent = new_parent
        self._fallback_to = new_parent

    @property
    def is_variant_set(self):
        """Return whether variant is set explicitly."""
        return self.is_attr_set(self.VARIANT_ATTR)

    @property
    def info(self):
        return {"name": self.name, "namespace": self.namespace, "variant": self.variant}

    @property
    def info_env(self):
        # MELTANO_EXTRACTOR_...
        return flatten({"meltano": {self.type.singular: self.info}}, "env_var")

    @property
    def all_commands(self):
        """Return a dictonary of supported commands, including those inherited from the parent plugin."""
        return {**self._parent.all_commands, **self.commands}

    @property
    def supported_commands(self):
        """Return all defined commands for the plugin."""
        return list(self.all_commands.keys())

    def env_prefixes(self, for_writing=False):
        prefixes = [self.name, self.namespace]

        if for_writing:
            prefixes.extend(self._parent.env_prefixes(for_writing=True))
            prefixes.append(f"meltano_{self.type.verb}"),  # MELTANO_EXTRACT_...

        return uniques_in(prefixes)

    @property
    def extra_config(self):
        return {f"_{k}": v for k, v in self.extras.items()}

    @property
    def config_with_extras(self):
        return {**self.config, **self.extra_config}

    @config_with_extras.setter
    def config_with_extras(self, new_config_with_extras):
        self.config.clear()
        self.extras.clear()

        for k, v in new_config_with_extras.items():
            if k.startswith("_"):
                self.extras[k[1:]] = v
            else:
                self.config[k] = v

    @property
    def settings(self):
        existing_settings = self._parent.settings
        return [
            *existing_settings,
            *SettingDefinition.from_missing(existing_settings, self.config),
        ]

    @property
    def extra_settings(self):
        existing_settings = self._parent.extra_settings
        return [
            *existing_settings,
            *SettingDefinition.from_missing(existing_settings, self.extra_config),
        ]

    @property
    def settings_with_extras(self):
        return [*self.settings, *self.extra_settings]

    def is_custom(self):
        return self.custom_definition is not None

    @property
    def is_shadowing(self):
        """Return whether this plugin is shadowing a base plugin with the same name."""
        return not self.inherit_from



import logging
import sys
import warnings

import click
import meltano
from meltano.core.behavior.versioned import IncompatibleVersionError
from meltano.core.logging import LEVELS, setup_logging
from meltano.core.project import Project, ProjectNotFound
from meltano.core.project_settings_service import ProjectSettingsService

logger = logging.getLogger(__name__)


@click.group(invoke_without_command=True, no_args_is_help=True)
@click.option("--log-level", type=click.Choice(LEVELS.keys()))
@click.option("-v", "--verbose", count=True)
@click.version_option(version=meltano.__version__, prog_name="meltano")
@click.pass_context
def cli(ctx, log_level, verbose):
    """
    Get help at https://www.meltano.com/docs/command-line-interface.html
    """
    if log_level:
        ProjectSettingsService.config_override["cli.log_level"] = log_level

    ctx.ensure_object(dict)
    ctx.obj["verbosity"] = verbose

    try:
        project = Project.find()
        setup_logging(project)

        readonly = ProjectSettingsService(project).get("project_readonly")
        if readonly:
            project.readonly = True

        if project.readonly:
            logger.debug("Project is read-only.")

        ctx.obj["project"] = project
    except ProjectNotFound as err:
        ctx.obj["project"] = None
    except IncompatibleVersionError as err:
        click.secho(
            "This Meltano project is incompatible with this version of `meltano`.",
            fg="yellow",
        )
        click.echo(
            "For more details, visit http://meltano.com/docs/installation.html#upgrading-meltano-version"
        )
        sys.exit(3)
